/************************************************************ //<>//
/* Author: Dr. Parson
/* Sketch: ShapePaintEcho2021NOGUI
/* ShapePaintEcho2021NOGUI February 2021 for Sentient AV retains MIDI, not Swing GUI.
/* ShapePaintEcho2019midi derives from ShapePaintEcho2018midi
/* during my spring 2019 semester sabbatical leave.
/* Revision dates January 2019:
/* Revision date: August 2018, updated from ShapePaintEcho2017
/* to send MIDI messages to Ableton Live's Max Surround Panner.
/* See sketch TestMaxSurroundPanner for background.
/* August 2018 version also has 'p' polar and 'c'  Cartesian motion.
/* Also the 'f' and 'F' snapshot brushes no longer paint halos.
/* Creation Date: 12/28/2016 through 4/1/2017, August-September 2017.
/* Professor Name: Dr. Parson
/* Sketch name: ShapePaintEcho2017, derived from ShapePaintSVG2017.
/* This sketch illustrates both interface inheritance (see
/* interface ShapeObjectTemplate) and implementation inheritance
/* (complete replacement of the display() method, and partial
/* replacement of some move() and shuffle() methods) in derived
/* classes. It also pulls a copy of the last un-erased pixels[]
/* at the start of draw(), stores them into PImage priorFrame,
/* and plots priorFrame (with optional rotate, scale, shearX and shearY)
/* as the background.
/* 12/29/2016: Concrete classes are named a through i,
/* as in commands +a, +i, -a, -i, etc.
/* 3/18/2017: Extend & adapt to accompany Ken Gehret at the 4/1/2017 gig.
/* 8/2017 add transformed echoes of past shapes in parallel to ability
/* to paint past images on display vi PImage morphing. Added +0 thru +9
/* and -0 thru -9 to get some more shapes.
/* August 26, 2017, added JFrameController, runs in another thread, make communication vars thread-safe
/* August 27, 2017, clean up command set and get close as possible match to PhotoMontage.pde.
/* April 22, 2018 added toggles 'O' erode and 'D' dilate filters
/* August 8-9, 2018 add 'c' for Cartesian move() default, 'p' for Polar around a circle.
 *********************************************************/
 
int midiDeviceIndex = 0 ;  // setup() checks for number of devices. Use one for output.

/********************************************************
 *
 * HOW TO INTERACT WITH THIS SKETCH VIA KEYS AFTER RUNNING
 
 +NAME TO ADD AN shape, where NAME is one of a through z, 0 through 9.
 from the list below. Adding makes it visible. You can also use
 A through Z to get a second object of the same class.
 
 -NAME TO REMOVE AN shape, where NAME is one of a through z, 0 through 9,
 from the list below. Removing makes it invisible.
 
 F to freeze the display movement, or to thaw if already frozen.
 
 M toggles tomove or ! tomove, toggling movement of SHAPES.
 
 f toggles fill off / on.
 
 V adds every possible shape; v removes all shapes.
 
 C to clear any unerased traces.
 
 R to clear display and remove all shapes (reset).
  
 DEPRECATED X|Y|Z to set background to black, yellow, cyan respectively.
  
 DEPRECATED: W|B to set shrink color to white, black respectively
 
 ~ to call shuffle() on all visible shapes.
 
 x[-]NUMBER[.]NUMBER\n translates from [0.0, 1.0] of width for repaint-last-frame translate (priorTranslateX)
 
 y[-]NUMBER[.]NUMBER\n translates from [0.0, 1.0] of height for repaint-last-frame translate (priorTranslateY)
 
 X[-]NUMBER[.]NUMBER\n shearX from -360 to 360 degrees (priorShearX)
 
 Y[-]NUMBER[.]NUMBER\n shearY from -360 to 360 degrees (priorShearY)
 
 WNUMBER.FRACTION\n wheels 0..16 for NUMBER, 0 or 1 is no wheel, FRACTION when non-0 scales overall display 1/fraction
 
 ZNUMBER\n where NUMBER is a fraction scaling each wheel size [0.1, 0.9] of overall.
 
 u[-]NUMBER[.]NUMBER\n shearX from -360 to 360 degrees (priorShearXSpeed)
 
 w[-]NUMBER[.]NUMBER\n shearY from -360 to 360 degrees (priorShearYSpeed)
 
 r[-]NUMBER[.]NUMBER\n rotates repaint-last-frame in degrees (priorRotate)
 
 s[-]NUMBER[.]NUMBER\n scales repaint-last-frame in float amouunt (priorScale)
 
 iNUMBER\n snapshots inner square for (0.0, 1.0], outer rectangular frame for (1.0, 2.0],
     uses circles instead of square for negative -NUMBER
 
 'BN\n' Set blendState to N, where N is an index into blendModes, with 0 being default BLEND mode.
 
 @NUMBER[.]NUMBER\n sets frameRate 1..60 (frameRate)
 
 eNUMBER\n sets eraseRate, where 0 does not erase prior frame, and 255 fully erases it.
 
 Enumber\n [-1, 400] where <= 0 is black, >= 360 white, others are hue for background().
 
 tNUMBER\n sets tailLength on history, where 0 is no tail, and limit is 15.
 
 gNUMBER\n sets tailGap between history steps, [1..50]
 
 PNUMBER\n sets posterize level 0..255
 
 TNUMBER\n sets threshold level 0.0..1.0
 
 ^NUMBER\n spawns children for each brush, where 2..6 spawns that many, 0 none, 1 removes all spawn
 
 G toggles gray monochrome
 O toggles erode filter
 D toggles dilate filter
 l melts out from a mask copy of image, melting from the darkest-to-lightest
 L melts out from a mask copy of image, melting from the lightest-to-darkest
 m melts in btoh directions at once
 control-i melts in both directions at once
 
 | flips toggles a horizontal flip about Y axis
 _ flips toggles a vertical flip about X axis
 { hashes pixels in X direction
 } hashes pixels in y direction
 / or \ resets both of these flips to no-flip state
 (The flips duplicate the original call to display() with a reflection.
 This happens to the object.display(), not to the PImage memory.
 It does show up in the next frame's PImage, of course.)
 
 mouseX supplies fluids rotate, mouseY does scale, bottom of display reverts back to scale == 1.
 Clicking mouse in lowest-left of display eliminates all rotation, scaling, shearing, translation.
 Clicking mouse anywhere onleft side of display eliminates shearing & translation.
 
 Dragging mouseX > 10 pixels sets global shearX; dragging mouseY does same for shearY.
 
 '#' resets priorTranslateX and priorTranslateY
 '[' resets priorTranslateX, ']' resets priorTranslateY, '(' resets priorRotate, ')' resets priorScale
 
 ********************************************************/

// See https://processing.org/reference/ConcurrentHashMap.html documentation.
// See http://docs.oracle.com/javase/7/docs/api/index.html java.util.ConcurrentHashMap
// for more details.
// For writing a p5.js Javascript versions that uses all of these features including
// implementation inheritance (there is no interface inheritance in a language
// that uses dynamic type) see: http://p5js.org/
// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Introduction_to_Object-Oriented_JavaScript
// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Inheritance_and_the_prototype_chain
// http://ccoenraets.github.io/es6-tutorial/


import java.util.concurrent.* ;
import java.util.List ;
import java.util.LinkedList ;
import java.util.Arrays ;
import java.util.Collections ; // wrap cross-thread LinkedList in a collections.synchronizedList
/***** START OF THREAD-SAFE VARS FOR JFrameController *****/
// team and visible are like arrays, except that they are indexed via
// Strings instead of ints. Use "put" to insert an ShapeObjectTemplate object
// into team or visible (indexed on its name), and "get" to retrieve
// an ShapeObjectTemplate element from team or visible.
// team holds all of the ShapeObjectTemplate objects.
// visible holds the visible ShapeObjectTemplate objects.
volatile String spath = null ;
final ConcurrentHashMap<String, ShapeData> team = new ConcurrentHashMap<String, ShapeData>();
final ConcurrentHashMap<String, ShapeData> visible = new ConcurrentHashMap<String, ShapeData>();
final ExecutorService threadpool = Executors.newCachedThreadPool();
volatile boolean firsttime = true ; // Use this the first time draw() runs.
volatile float priorRotate = 0.0;  // Used to rotate prior frame when != 0.0
volatile float priorScale = 1.0 ; // Used to scale prior frame when != 1.0
volatile int priorTranslateX = 0, priorTranslateY = 0 ; // Used to translate prior frame when != 0.0, 'X' and 'Y'
volatile float priorShearX = 0, priorShearY = 0 ; // Used to shear prior frame when != 0.0, in degrees, 'X' and 'Y'
volatile float priorShearXSpeed = 0, priorShearYSpeed = 0 ; // 'u' and 'w'
volatile int flipx = 1, flipy = 1 ; // flipx = -1 flips drawing coordinates about the Y axis, for example, via scale
volatile boolean hashX = false, hashY = false ;
volatile int hashXOffset = 0, hashYOffset = 0 ;
volatile float wheelCount = 1.0 ; // 'W' wheels within wheels, 1/28/2019
volatile float wheelScale = 0.5 ; // 'Z' scale for size of wheelCount extra wheels
// volatile int reflectx = 0, reflecty = 0 ; // seconds 1..60 or 0 for none to reflect a shape
volatile boolean tomove = true ; // 'M' toggles this to bypass move() calls on shapes.
volatile boolean isClear = false ;  // Set to true to clear old traces, toggle via "C"
volatile boolean isReset = false ;
volatile boolean isExit = false ;
volatile boolean isFrozen = false ; // Set to true to freeze the display, toggle via "F"
volatile boolean isPolar = false ; // true for isCartesian
volatile int eraseRate = 255 ; // Speed of erasure, 0 to 255, in keyPressed().
volatile boolean showfill = true, showstroke = true ;
volatile int tailGap = 10 ;
volatile int tailLength = 0 ;
class HistoryContainer {
  final List<ShapeData> objects = Collections.synchronizedList(new LinkedList<ShapeData>());
}
final List<HistoryContainer> history = Collections.synchronizedList(new LinkedList<HistoryContainer>());
// START OF NON-VOLATILE OBJECT HISTORY (instead of CANVAS HISTORY) fields:
int historyPhase = 0 ; // We clone on 1/tailGap states.
// clippingCircle hides everything outside a circle centered in the display;
// it is a black PNG square that is transparent in its central circle.
PImage clippingCircle ;
// Everything is centered in the display. See initialization at top of draw().
int domeTranslateX, domeTranslateY, domeWidthHeight ;
String command = ""; // Build + and - command strings in this String.
PImage priorFrame = null ; // Used to copy pixels from prior frame when transforming prior frame image
int pressedX = 0, pressedY = 0 ; // save mousePressed coords in mousePressed(), use in mouseReleased()
color shrinkColor = color(0);
final color black = color(0);
int leftClippingBound, rightClippingBound ; // Used in per-pixel clipping of prior frame for a Mac bug.
float oneSecondSpeed ; // See draw() for speeds
float speedlow ;
float speedmid ;
float speedhigh ;
final float SCALELIMIT = 10.0 ;
int spawnDegree = 0 ; // added Jan2019 for shapes that can recur.

int eraseColor = color(0);
int lastEraseColor = eraseColor ;
boolean isgray = false, iserode = false, isdilate = false,
  isinvert = false ;
int posterizeLevel = 0;
float thresholdLevel = 0.0 ;

int blendModes [] = {
  BLEND, ADD, SUBTRACT, LIGHTEST, DARKEST,  
  SCREEN, MULTIPLY, EXCLUSION, DIFFERENCE, REPLACE
} ; 
int blendState = 0 ;
float nextcapture = 0.0, lastcapture = 0.0 ;  // Added Jan2019 for screen photography
char captureShape = 'i' ;  // i for circle, I for square
PImage captureimg = null ;
int meltindex = 0 ; // 2 for initial 'l' and -2 for 'L', drop to 1 or -1 when in process, 0 when not used
int meltlowpixel = 0 ; // lowest pixel value used for melting
int melthipixel = 0 ; // highest pixel in bidirectional melting on command 'm'
// a melt uses the captureimg, and maintains a grayscale meltmaskErode as well
PImage meltmaskErode = null ;
PImage meltmaskDilate = null ;  // used for bidirectional melting
boolean meltBothWays = false ;

// MIDI STUFF
import javax.sound.midi.* ;  // Get everything in the Java MIDI (Musical Instrument Digital Interface) package.
// See http://midi.teragonaudio.com/tech/midispec.htm for CONTROLLER codes
final int CTRLNUM = 0 ; // CTRLNUM+0 through CTRLNUM+5 used for Max Surround Panner
final int CONTROLLER = 0xb0 ; // MIDI Controller opcode
// MIDI VARIABLES, see http://faculty.kutztown.edu/parson/spring2017/MidiKBSpring2016Parson.txt

// NOTE: A final variable is in fact a constant that cannot be changed.
MidiDevice.Info[] midiDeviceInfo = null ;
// See javax.sound.midi.MidiSystem and javax.sound.midi.MidiDevice
MidiDevice device = null ;
// See javax.sound.midi.MidiSystem and javax.sound.midi.MidiDevice
Receiver receiver = null ;
// javax.sound.midi.Receiver receives your OUTPUT MIDI messages (counterintuitive?)
// SEE https://www.midi.org/specifications/item/gm-level-1-sound-set but start at 0, not 1
final int NUMINSTRUMENTS = 7 ; // in Ableton Live

void displayHistory() {
  // println("DEBUG displayHistory");
  if (history != null && (tailLength > 0)) {
    int limit = constrain(tailLength, 0, 15);
    while (history.size() > limit) {
      HistoryContainer then = history.get(0);
      for (ShapeData thingy : then.objects) {
        thingy.brightness = 0;
        thingy.display();
      }
      history.remove(0);
    }
  }
  if (history != null && history.size() > 0 && (tailLength > 0)) {
    for (int i = 0; i < history.size(); i++ ) {
      HistoryContainer then = history.get(i);
      for (ShapeData thingy : then.objects) {
        thingy.display();
      }
    }
  }
}

void setup() {
  size(1900, 1080, P2D);
  frameRate(30);  // avoid mac bug 11/2021
  // MIDI:
  // 1. FIND OUT WHAT MIDI DEVICES ARE AVAILABLE FOR VARIABLE midiDeviceIndex.
  int i = 0 ;
  midiDeviceInfo = MidiSystem.getMidiDeviceInfo();
  for (i = 0; i < midiDeviceInfo.length; i++) {
    println("MIDI DEVICE NUMBER " + i + " Name: " + midiDeviceInfo[i].getName()
      + ", Vendor: " + midiDeviceInfo[i].getVendor()
      + ", Description: " + midiDeviceInfo[i].getDescription());
  }
  // 2. OPEN ONE OF THE MIDI DEVICES UP FOR OUTPUT.
  try {
    Exception rx = null ;
    if (midiDeviceIndex >= 0) {
      device = MidiSystem.getMidiDevice(midiDeviceInfo[midiDeviceIndex]);
      device.open();  // Make sure to close it before this sketch terminates!!!
      // There should be a way to schedule a method when Processing closes this
      // sketch, so we can close the device there, but it is not documented for Processing 3.
      try {
        receiver = device.getReceiver();
      } catch (Exception rxx) {
        rx = rxx ;
      }
    }
    if (receiver == null && midiDeviceIndex >= 0) {
      println("ERROR, MIDI midiDeviceIndex " + midiDeviceIndex + " does not have a receiver "
        + ((rx == null) ? "" : rx.getMessage()));
    }
    // NOTE: Either of the above method calls can throw MidiUnavailableException
    // if there is no available device or if it does not have a Receiver to
    // which we can send messages. The catch clause intercepts those error messages.
    // See https://www.midi.org/specifications/item/gm-level-1-sound-set, use voice variable
    /*
    ShortMessage noteMessage = new ShortMessage() ;
     noteMessage.setMessage(ShortMessage.PROGRAM_CHANGE, 0, voice, 0); // to channel 0
     receiver.send(noteMessage, -1L);  // send it now
     */
  } 
  catch (MidiUnavailableException mx) {
    println("MIDI UNAVAILABLE " + mx.getMessage()); // Error messages go here.
    device = null ;
    receiver = null ; // Do not try to use them.
    threadpool.shutdown();
    exit();
  } catch (ArrayIndexOutOfBoundsException aix) {
    println("ArrayIndexOutOfBoundsException, Invalid midiDeviceIndex " + midiDeviceIndex + ", " + aix.getMessage());
    /*
  } catch (InvalidMidiDataException dx) {
     println("MIDI ERROR: " + dx.getMessage()); // Error messages go here.
     */
  }
  // ORIGINAL, PRE-MIDI STARTUP CODE FOLLOWS:
  spath = sketchPath("");
  //println("sketchPath is " + spath + ", dataPath is " + dataPath(""));
  // fullScreen(P2D); // stick with 2D so that the clippingCircle works correctly.
  domeTranslateX = width/2 ; // plot everything relative to center of dome
  domeTranslateY = height/2 ;
  background(eraseColor); // Initially the planetarium background is all black.
  
  // eraseRate = 0 ; // Tweak frameRate & eraseRate for musical instrument classes s through z
  clippingCircle = loadImage("CenteredClippingCircle.png");
  // Put one keyed object for each class into team.

  char letter = 'a' ;
  i = 0 ;
  for (; letter <= 'z'
    ; i++, letter++) {  // '?' through 'z'
    String myname = (new Character(letter)).toString() ;
    team.put(myname, 
      new ShapeDrawer(myname,(int)random(0, 360), i, radians(-3.0), 
      0.01, 0.015, 0.0, 0.0, true));
    myname = (new Character(Character.toUpperCase(letter))).toString();
    team.put(myname, 
      new ShapeDrawer(myname, (int)random(0, 360), i, radians(-2.0), 
      0.02, 0.025, ((i & 1) == 1) ? radians((i+1) * 0.125) : 0, 
      ((i & 1) == 0) ? radians(-0.25 * (i+1)): 0, true));
  }
  for (letter = '0'; letter <= '9'
    ; i++, letter++) {
    String myname = (new Character(letter)).toString() ;
    team.put(myname, 
      new ShapeDrawer(myname,(int)random(0, 360), i, radians(-3.0), 
      0.01, 0.015, 0.0, 0.0, true));
    /*team.put((new Character(Character.toUpperCase(letter))).toString(),
     new ShapeDrawer((int)random(0,360), i, radians(-2.0),
     0.02, 0.025, ((i & 1) == 1) ? radians((i+1) * 0.125) : 0,
     ((i & 1) == 0) ? radians(-0.25 * (i+1)): 0, true));*/
  }
  //frameRate(30); // frameRate(60);
}

void draw() {
  // println("sketchPath is " + sketchPath("") + ", dataPath is " + dataPath(""));
  if (isExit) {
    threadpool.shutdown();
    exit();
  }
  if (isFrozen) {
    return ;
  }
  blendMode(BLEND);
  colorMode(RGB, 255, 255, 255, 255); // default. Shape objects may use other colorModes.
  if (eraseRate >= 255 || eraseColor != lastEraseColor) {
    background(eraseColor);
    lastEraseColor = eraseColor ;
  }
  if (firsttime) {
    // This can't be in setup() because we need width & height values.
    firsttime = false ;  // Do not come in here a second time.
    initEraseLocs(width, height);
    domeTranslateX = width/2 ; // plot everything relative to center of dome
    domeTranslateY = height/2 ;
    domeWidthHeight = min(width, height);
    priorFrame = createImage(width, height, RGB);
    leftClippingBound = (width - domeWidthHeight) / 2 ;
    rightClippingBound = width - leftClippingBound ;
    oneSecondSpeed = (float) height / (float) frameRate ;
    speedlow = oneSecondSpeed / 20.0 ; // 30 seconds across
    speedmid = oneSecondSpeed / 10.0 ; // 15 seconds across
    speedhigh = oneSecondSpeed / 5.0 ; // 5 seconds across
  }
  if (isClear) {
    background(eraseColor);
    isClear = false ;
  }
  if (isReset) {
    background(0);
    isReset = false ;
    eraseColor = 0 ;
  }

  if (priorRotate != 0.0 || priorScale != 1.0 || priorTranslateX != 0 || priorTranslateY != 0
    || priorShearX != 0.0 || priorShearY != 0.0
    || priorShearXSpeed != 0.0 || priorShearYSpeed != 0.0
    /* || flipx != 1 || flipy != 1 */) {
    // 0.0 and 1.0 are exact in IEEE floating point.
    // (priorRotate != 0.0 || priorScale != 1.0 || priorTranslateX != 0 || priorTranslateY != 0)
    morphCanvas();
  } 

  // The next commands through rect cause partial erasure of whole display.
  colorMode(RGB, 255, 255, 255, 255);
  //fill(0, 0, 0, eraseRate);
  //stroke(0, 0, 0, eraseRate);
  /*
  fill(eraseColor, eraseRate);
   stroke(eraseColor, eraseRate);
   rectMode(CORNER);
   rect(0, 0, width, height); // Paint a transparent, translucent, or opaque mask over entire 2D display.
   */
  applyEraseLocs(eraseColor, eraseRate);

  // This section of code display()s and move()s the shape objects in map visible:
  pushMatrix();
  // Center 0,0 at middle of display (top of the dome).
  translate(domeTranslateX, domeTranslateY);
  strokeWeight(1); // Global weight adjusted by up & down arrows on keyboard.
  colorMode(RGB, 255, 255, 255, 255);
  HistoryContainer now = null ;
  if ((tailLength > 0) && (historyPhase % tailGap) == 0) {
    now = new HistoryContainer();
  }
  if ((tailLength > 0) && history.size() > 0) {
    displayHistory();
  }
  historyPhase = (historyPhase+1) % tailGap ;
  blendMode(blendModes[blendState]);
  // pushMatrix();  // wrap wheelCount transform
  int iwc = int(wheelCount);
  float fwc = wheelCount - iwc ;  // get the fractional part of wheelCount
  if (fwc > 0.0 && fwc < 1.0) {
    float wc = 1.0 / fwc ;
    scale(wc);
  }
  for (ShapeData thingy : visible.values()) {
    colorMode(RGB, 255, 255, 255, 255); // In case a shape object changes colorMode.
    noFill();  // Just in case display() sets a fill().
    noStroke(); // Just in case display() sets a stroke().
    noTint();  // Just in case display() sets a tint().
    strokeWeight(2);
    // Display and move each visible shape.
    thingy.display();
    if (flipy == -1) {
      pushMatrix();
      scale(1.0, -1.0);
      thingy.display();
      popMatrix();
      // no cloning of the reflections for now
    }
    if (flipx == -1) {
      pushMatrix();
      scale(-1.0, 1.0);
      thingy.display();
      popMatrix();
      // no cloning of the reflections for now
    }
    if (flipx == -1 && flipy == -1) {
      pushMatrix();
      scale(-1.0, -1.0);
      thingy.display();
      popMatrix();
      // no cloning of the reflections for now
    }
    if (now != null) {
      ShapeData tmp = (ShapeData)thingy.clone();
      now.objects.add(tmp);
    }
    if (tomove) {
      thingy.move();
    }
  }
  spawnDegree = 0 ;  // If > 1 it was handled in move(), now reset.
  if (now != null) {
    //println("DEBUG add to history ");
    history.add(now);
  }
  blendMode(BLEND);
  if (isinvert) {
    filter(INVERT);
    isinvert = false ;  // avoid framerate flicker
  }
  if (iserode) {
    filter(ERODE);
  }
  if (isdilate) {
    filter(DILATE);
  }
  if (posterizeLevel > 1) {
    filter(POSTERIZE, posterizeLevel);
  }
  if (isgray) {
    filter(GRAY);
  }
  if (thresholdLevel > 0.0 && thresholdLevel < 1.0) {
    filter(THRESHOLD, thresholdLevel);
  }
  if (hashX || hashY) {
    int leftmarg, rightmarg, topmarg, botmarg ;
    if (width >= height) {
      topmarg = 0 ;
      botmarg = height ;
      leftmarg = (width-height) / 2;
      rightmarg = leftmarg + height ;
    } else {
      leftmarg = 0 ;
      rightmarg = width ;
      topmarg = (height-width) / 2;
      botmarg = topmarg + width ;
    }
    int diam = rightmarg - leftmarg ;
    loadPixels();
    PImage backup = createImage(width, height, ARGB);
    backup.loadPixels();
    if (hashX) {
      for (int y = topmarg ; y < botmarg ; y++) {
        for (int x = leftmarg ; x < rightmarg ; x++) {
          float yp = float(y-topmarg) / float(diam);
          int newx = (round(yp * diam + (x-leftmarg)) % diam) + leftmarg ;
          backup.pixels[y*width+newx] = pixels[y*width+x];
        }
      }
      System.arraycopy(backup.pixels, 0, pixels, 0, pixels.length);
    }
    if (hashY) {
      for (int y = topmarg ; y < botmarg ; y++) {
        for (int x = leftmarg ; x < rightmarg ; x++) {
          float xp = float(x-leftmarg) / float(diam);
          int newy = (round(xp * diam + (y-topmarg)) % diam) + topmarg ;
          backup.pixels[newy*width+x] = pixels[y*width+x];
        }
      }
      System.arraycopy(backup.pixels, 0, pixels, 0, pixels.length);
    }
    updatePixels();
  }
  noTint();
  if (meltindex != 0) {
    nextcapture = lastcapture = 0 ;
    if (meltindex < -1 || meltindex > 1) {
      // capture a new meltmaskErode after one last plot of former captureimg
      if (captureimg != null) {
        // Put the prior one on screen before re-capturing.
        imageMode(CENTER);
        image(captureimg, 0,0);
      }
      captureimg = createImage(width, height, ARGB);
      loadPixels();
      captureimg.loadPixels();
      System.arraycopy(pixels, 0, captureimg.pixels, 0, pixels.length);
      captureimg.updatePixels();
      meltmaskErode = captureimg.copy();
      meltmaskErode.filter(GRAY);
      if (meltindex < -1) {
        meltmaskErode.filter(INVERT);
      }
      if (meltBothWays) {
        // println("DEBUG MELT BOTH WAYS");
        meltmaskDilate = captureimg.copy();
        meltmaskDilate.filter(GRAY);
        meltmaskDilate.loadPixels();
        melthipixel = 0 ;
        for (int pix : meltmaskDilate.pixels) {
          if ((pix&0x0ff) > melthipixel) {
            melthipixel = (pix&0x0ff) ;
          }
        }
        meltmaskDilate.updatePixels();
      } else {
        meltmaskDilate = null ;
      }
      meltlowpixel = 256 ;
      meltmaskErode.loadPixels();
      for (int pix : meltmaskErode.pixels) {
        if ((pix&0x0ff) < meltlowpixel) {
          meltlowpixel = (pix&0x0ff) ;
        }
      }
      meltmaskErode.updatePixels();
      meltindex = (meltindex < -1) ? -1 : 1 ;
    }
    meltmaskErode.filter(ERODE);
    meltmaskErode.loadPixels();
    if (meltmaskDilate != null) {
      meltmaskDilate.filter(DILATE);
      meltmaskDilate.loadPixels();
    }
    captureimg.loadPixels();
    boolean keepgoing = false ;
    for (int pindex = 0 ; pindex < meltmaskErode.pixels.length ; pindex++) {
      if (((meltmaskErode.pixels[pindex]&0x0ff) <= meltlowpixel)
          || (meltmaskDilate != null && (meltmaskDilate.pixels[pindex]&0x0ff) >= melthipixel)) {
        // println("DEBUG LOW PIX");
        if (captureimg.pixels[pindex] != 0) {
          keepgoing = true ;
          captureimg.pixels[pindex] = 0 ;
          // meltmaskErode.pixels[pindex] = 0xff000000 ;
          // println("DEBUG NEW LOW PIX");
        }
      } else {
        keepgoing = true ;
      }
    }
    meltmaskErode.updatePixels();
    if (meltmaskDilate != null) {
      meltmaskDilate.updatePixels();
    }
    captureimg.updatePixels();
    if (! keepgoing) {
      captureimg = meltmaskErode = meltmaskDilate = null ;
      meltindex = meltlowpixel = 0 ;
      // println("DEBUG NO KEEP GOING");
    } else {
      imageMode(CENTER);
      image(captureimg, 0,0);
      // image(meltmaskErode,0,0); // DEBUG
      // println("DEBUG YES KEEP GOING");
    }
  } else if (nextcapture != lastcapture) {
    if (nextcapture == 0.0) {
      captureimg = null ;
    } else {
      if (captureimg != null) {
        // Put the prior one on screen before re-capturing.
        imageMode(CENTER);
        image(captureimg, 0,0);
      }
      captureimg = createImage(width, height, ARGB);
      PImage partialmask = createImage(width, height, ARGB);
      meltmaskErode = null ;
      meltmaskDilate = null ;
      meltindex = meltlowpixel = 0 ;
      loadPixels();
      captureimg.loadPixels();
      partialmask.loadPixels();
      System.arraycopy(pixels, 0, captureimg.pixels, 0, pixels.length);
      Arrays.fill(partialmask.pixels, 0); // transparent until we say otherwise
      int xcenter = width / 2 ;
      int ycenter = height / 2 ;
      int radius = min(width/2, height/2) ;
      float ratio = radius / ((abs(nextcapture))) ;
      if (captureShape == 'i') {
        for (int y = 0, ynorm = 0 ; ynorm < height ; y+= width, ynorm++) {
          for (int x = 0 ; x < width ; x++) {
            float steps = (dist(x,ynorm,xcenter,ycenter));
            int region = round(steps / ratio);
            if (((region & 1) == 0 && nextcapture > 0) || ((region & 1) == 1 && nextcapture < 0)) {
              partialmask.pixels[y+x] = 0xffffffff ;
            }
          }
        }
      } else {
        for (int y = 0, ynorm = 0 ; ynorm < height ; y+= width, ynorm++) {
          for (int x = 0 ; x < width ; x++) {
            int steps = abs(x-xcenter) + abs(ynorm-ycenter) ; // Manhatten distance
            int region = round(steps / ratio);
            if (((region & 1) == 0 && nextcapture > 0) || ((region & 1) == 1 && nextcapture < 0)) {
              partialmask.pixels[y+x] = 0xffffffff ;
            }
          }
        }
      }
      partialmask.updatePixels();
      captureimg.updatePixels();
      partialmask.filter(GRAY);
      captureimg.mask(partialmask);
      /*
      float normcapture = (abs(nextcapture) <= 1.0) ? abs(nextcapture) : 
        abs(nextcapture) - 1.0;
      float nextlose = 1.0 - normcapture ; // abs(nextcapture) ;
      boolean innerphoto = (abs(nextcapture) <= 1.0);
      float xmargin = (nextlose * width) / 2.0 ;
      float ymargin = (nextlose * height) / 2.0 ;
      if (nextcapture > 0.0) {   
        // if (nextcapture < 1.0) {
          if (width >= height) {
            xmargin += nextlose * (width-height) / 2.0 ;
          } else {
            ymargin += nextlose * (height-width) / 2.0 ;
          }
          carveSquarePixels(captureimg.pixels, width, height, round(xmargin), round(ymargin), 
            round(width-xmargin), round(height-ymargin), innerphoto);
        // }
      } else {
        // negative
        carveCirclePixels(captureimg.pixels, width, height, round(xmargin), round(ymargin), 
            round(width-xmargin), round(height-ymargin), innerphoto);
      }
      captureimg.updatePixels();
      */
    }
    lastcapture = nextcapture ;
  }
  if (captureimg != null) {
    imageMode(CENTER);
    image(captureimg, 0, 0);
  }
  // popMatrix();  // unwrap wheelCount transform before clipping
  int wcnt = round(wheelCount);
  if (wcnt >= 2) {
    PImage little = createImage(width, height, ARGB);
    loadPixels();
    little.loadPixels();
    System.arraycopy(pixels, 0, little.pixels, 0, pixels.length);
    int wid, hgt,  xleft, ytop, xright, ybot ;
    if (width >= height) {
      hgt = height ;
      wid = width ;
      xleft = (width-height)/2 ;
      xright = xleft + wid ;
      ytop = 0 ;
      ybot = height-1 ;
    } else {
      wid = width ;
      hgt = height ;
      xleft = 0 ;
      xright = width-1 ;
      ytop = (height-width)/2 ;
      ybot = ytop + hgt ;
    }
    carveCirclePixels(little.pixels,wid,hgt,xleft,ytop,xright,ybot,true);
    little.updatePixels();
    updatePixels();
    little.resize(round(width * wheelScale), round(height * wheelScale));
    pushMatrix();
    //background(eraseColor);
    // translate(width/4, 0);
    float roter = TWO_PI / wcnt ;
    for (int i = 0 ; i < wcnt ; i++) {
      image(little,width/8, 0);
      rotate(roter);
    }
    popMatrix();
  }
  colorMode(RGB, 255, 255, 255, 255);
  noFill();  // Just in case display() sets a fill().
  noStroke(); // Just in case display() sets a stroke().
  noTint();  // Just in case display() sets a tint().
  // Clip the center area using the clippingCircle PNG image,
  // then clip the rectangular area around it using clip().
  imageMode(CENTER);
  image(clippingCircle, 0, 0, domeWidthHeight, domeWidthHeight);
  clip(width/2, height/2, domeWidthHeight, domeWidthHeight);
  popMatrix();
} // end draw() function

void carveSquarePixels(int [] cpixels, int wid, int hgt, int xleft, int ytop,
    int xright, int ybot, boolean innerphoto) {
  if (ybot >= hgt) {
    ybot = hgt-1;
  }
  if (innerphoto) {
    if (ytop > 0) {
      Arrays.fill(cpixels, 0, ytop * wid, 0);
    }
    if ((ybot+1) < hgt) {
      Arrays.fill(cpixels, (ybot+1) * wid, cpixels.length, 0);
    }
    for (int y = ytop * wid; y <= ybot * wid; y += wid) {
      for (int x = 0; x < wid; x++) {
        if (x < xleft || x > xright) {
          // Jan2019 make it square
          cpixels[y+x] = 0 ;
        }
      }
    }
  } else {
    // println("ENTER BARREN DEBUG " + xleft + "," + ytop + " TO " + xright + "," + ybot);
    for (int y = 0, ynorm = 0 ; y < cpixels.length ; y += wid, ynorm++) {
      for (int x = 0; x < wid; x++) {
        if (x > xleft && x < xright && ynorm > ytop && ynorm < ybot) {
          // println("DEBUG EMPTY INNER PIXEL");
          cpixels[y+x] = 0 ;
        }
      }
    }
  }
}
void carveCirclePixels(int [] cpixels, int wid, int hgt, int xleft, int ytop,
    int xright, int ybot, boolean innerphoto) {
  // Carve out the square that contains this circle, then iterate over circle ion middle of the square.
  if (ybot >= hgt) {
    ybot = hgt-1;
  }
  int xcenter = wid / 2 ;
  int ycenter = hgt / 2 ;
  int xrad = xright - xleft ;
  int yrad = ybot - ytop ;
  int radius = ((yrad <= xrad) ? yrad : xrad) / 2 ;
  if (innerphoto) {
    if (ytop > 0) {
      Arrays.fill(cpixels, 0, ytop * wid, 0);
    }
    if ((ybot+1) < hgt) {
      Arrays.fill(cpixels, (ybot+1) * wid, cpixels.length, 0);
    }
    for (int y = ytop * wid, ynorm = ytop; y <= ybot * wid; y += wid, ynorm++) {
      for (int x = 0; x < wid; x++) {
        if (dist(x,ynorm,xcenter,ycenter) >= radius) {
          cpixels[y+x] = 0 ;
        }
      }
    }
  } else {
    for (int y = 0, ynorm = 0; y < cpixels.length; y += wid, ynorm++) {
      for (int x = 0; x < wid; x++) {
        if (dist(x,ynorm,xcenter,ycenter) < radius) {
          cpixels[y+x] = 0 ;
        }
      }
    }
  }
}

void morphCanvas() {
  loadPixels() ; // get pixels from last display frame into pixels[]
  priorFrame.loadPixels(); // transfer them (below) into the priorFrame PImage for modified-redisplay.
  int lim = pixels.length;
  if (priorFrame.pixels.length < lim) { // They should have identical length.
    lim = priorFrame.pixels.length;
  }
  // println("pixels.length is " + pixels.length + ", priorFrame.pixels.length is " + priorFrame.pixels.length) ;
  int xp = 0; // , yp = 0 ; The pixel layout is a horizontal row at a time, from top-to-bottom. xp is the row.
  if (priorScale < 1.0) {
    // Mac copies some off-display pixels into the image boundary when scaling down, so avoid that
    // by setting explicit shrinkColor pixels into the boundary instead.
    // For Javascript pay attention to pixel density, see code in:
    // http://p5js.org/reference/#/p5/pixels[]
    for (int i = 0; i < lim; i++) {
      if (xp <= leftClippingBound || xp >= rightClippingBound) {
        priorFrame.pixels[i] = shrinkColor ;
      } else {
        priorFrame.pixels[i] = pixels[i];
      }
      if (++xp >= width) {
        xp = 0 ;
        // DON'T NEED yp: yp++ ;
      }
    }
  } else {
    System.arraycopy(pixels, 0, priorFrame.pixels, 0, lim);
    /*
    for (int i = 0; i < lim; i++) { // java.util.Arrays is missing this capability!
      priorFrame.pixels[i] = pixels[i];
    }
    */
  }
  priorFrame.updatePixels(); // Put the pixels into the priorFrame PImage.
  imageMode(CENTER);
  pushMatrix();
  // Everything up to next popMatrix() is for re-painting the prior frame.
  if (priorTranslateX != 0 || priorTranslateY != 0) {
    // println("DEBUG translating " + priorTranslateX + "," + priorTranslateY);
    int xxx = (priorTranslateX == 0) ? width/2 : (width / 2 + priorTranslateX) % width ;
    int yyy = (priorTranslateY == 0) ? height/2 : (height / 2 + priorTranslateY) % height ;
    translate(xxx, yyy);
  } else {
    // println("DEBUG translating standard");
    // Every coordinate is relative to center of the dome.
    translate(width/2, height/2);
  }
  if (priorRotate != 0.0) {
    rotate(priorRotate);
  }
  if (priorScale != 1.0) {
    scale(priorScale);
  }
  priorShearX += priorShearXSpeed ;
  while (priorShearX <= -360) {
    priorShearX += 360 ;
  }
  while (priorShearX >= 360) {
    priorShearX -= 360 ;
  }
  priorShearY += priorShearYSpeed ;
  while (priorShearY <= -360) {
    priorShearY += 360 ;
  }
  while (priorShearY >= 360) {
    priorShearY -= 360 ;
  }
  if (priorShearX != 0.0) {
    shearX(radians(priorShearX));
  }
  if (priorShearY != 0.0) {
    shearY(radians(priorShearY));
  }
  /*
    if (flipx != 1 || flipy != 1) {
   scale(flipx, flipy);
   }
   */
  background(eraseColor);
  image(priorFrame, 0, 0);
  popMatrix();
}

/** 
 *  A Processing interface is a specification for one or more classes of objects.
 *  ShapeObjectTemplate is a specification for one or more shape object classes.
 *  You can add new ShapeObjectTemplate-derived classes. Mine are below.
 *  You must implement each of the functions in your class according to the
 *  assignment handout.
 **/
interface ShapeObjectTemplate extends java.lang.Cloneable {
  /**
   *  Your ShapeObjectTemplate-derived class must have one or more variable
   *  data fields, at a minimum "int x" and "int y" for the x,y location,
   *  where 0,0 is the middle of the display area.
   *  Write this in your shape class, not here in ShapeObjectTemplate.
   **/
  /** You should write a constructor that takes 0 parameters. **/
  /**
   *  Write a display function that starts like this:
   
   pushMatrix();
   translate(x, y);
   
   and ends like this:
   
   popMatrix();
   
   with all display code inside the function.
   You are not required to display an image.
   Write this in your shape class, not here in ShapeObjectTemplate.
   
   In addition to translate, the display() code in your class must use
   one or more of scale (with 1 or 2 arguments), rotate,
   shearX, or shearY. You can also manipulate variables for color & speed.
   See my example classes for ideas. You can use an image, with
   or without tint, but that is not required. If you use tint, please call
   noTint() before the end of display().
   **/
  void display();
  /** Write move() to update variable fields inside the object as before.
   *  Write this in your shape class, not here in ShapeObjectTemplate. **/
  void move();
  /** Write shuffle() to randomize some of the variable fields inside the object.
   *  Write this in your shape class, not here in ShapeObjectTemplate. **/
  void shuffle();
  /** clone() clones the object state **/
  public ShapeObjectTemplate clone();
}

/**
 *  ShapeDataHelper holds the data fields used by all the shapes.
 *  It also implements the basic move() method to keep the shap on the screen.
 *  It implements only the parts of shuffle() that overlap with its move().
 *  It uses floats for locatio & speed to accommodate slow moves at 60 fps.
 *  It assumes HSB color mode. Saturation, brightness, and alpha are
 *  always always 100% except for black and white.
 **/
abstract class ShapeData implements ShapeObjectTemplate {
  protected String myname = "unnamed" ;
  protected int spawnSerialNo = 0 ;
  protected float x, y, speedx, speedy ;
  protected int hue, saturation, brightness ;
  // Some subclass may not use the next fields, but I am
  // defining them here to avoid scattering them across the
  // design, & to support prototyping them in & out of subclasses.
  // Java initializes them to zeroes by default.
  protected float scalex = 1.0, scaley = 1.0 ;
  protected float scalexSpeed, scaleySpeed ;
  protected float rotateRadians, shearxRadians, shearyRadians ;
  protected float rotateSpeedRadians, shearxSpeedRadians, 
    shearySpeedRadians ;
  protected boolean IamAclone = false ; // added 2/17/2018
  protected float myreflxamount = 1.0, myreflyamount = 1.0 ;
  void clonehelp(ShapeData newone) {
    // DO NOT CLONE SPEED FIELDS, CLONES DO NOT NEED THEM.
    newone.hue = hue ;
    newone.saturation = saturation ;
    newone.brightness = brightness ;
    newone.x = x ;
    newone.y = y ;
    //newone.speedx = speedx ;
    //newone.speedy = speedy ;
    newone.scalex = scalex ;
    newone.scaley = scaley ;
    //newone.scalexSpeed = scalexSpeed ;
    //newone.scaleySpeed = scaleySpeed ;
    newone.rotateRadians = rotateRadians ;
    newone.shearxRadians = shearxRadians ;
    newone.shearyRadians = shearyRadians ;
    //newone.rotateSpeedRadians = rotateSpeedRadians ;
    //newone.shearxSpeedRadians = shearxSpeedRadians ;
    //newone.shearySpeedRadians = shearySpeedRadians ;
    newone.IamAclone = true ; // added 2/17/2018
  }
  public ShapeData clone() {
    return null ;
  }
  /**
   *  Use 0 <= myhue <= 359 for HSB with saturation & brightness of 100%.
   *  Use myhue > 359 for white, < 0 for black.
   *  Does not call shuffle().
   **/
  public ShapeData(String thename, int myhue) {
    myname = thename ;
    if (myhue < 0) {
      hue = saturation = brightness = 0 ;
    } else if (myhue > 359) {
      hue = saturation = 0 ;
      brightness = 99 ;
    } else {
      hue = myhue ;
      saturation = brightness = 99 ;
    }
  }
  /** shuffle() sets x, y, speedx, speedy **/
  public void shuffle() {
    float lowspeed, highspeed ;
    if (speedx == 0 && speedy == 0) {
      // start slowly
      lowspeed = 0 ;
      highspeed = speedlow ;
    } else if (speedx < speedlow || speedy < speedlow) {
      lowspeed = speedlow ;
      highspeed = speedmid ;
    } else if (speedx < speedmid || speedy < speedmid) {
      lowspeed = speedmid ;
      highspeed = speedhigh ;
    } else {
      lowspeed = 0 ; 
      ;
      highspeed = speedlow ;
    }
    speedx = random(lowspeed, highspeed);
    speedy = random(lowspeed, highspeed);
    x = (random(-domeWidthHeight/2, domeWidthHeight/2));
    y = (random(-domeWidthHeight/2, domeWidthHeight/2));
  }
  /**
   *  Move x and y according to speedx and speedy.
   *  Maintain these as floats for slow moves.
   *  Move scale, rotate, and shear fields when they have non-0
   *  speed, limit angles to (-TWO_PI, TWO_PI) radians,
   *  limit scale to (-SCALELIMIT, SCALELIMIT).
   **/
  public void move() {
    float radius = min(width/2, height/2);
    if (isPolar) {
      float normx = x / radius ;
      float normy = (-y) / radius ;
      float [] radiusThenAngle = cartesianToPolar(normx, normy);
      float band = radiusThenAngle[0] ; // 0.0 to 0.33, .331 to .67, .671 to 1.0
      radiusThenAngle[0] += (speedx / radius) / 3.0 ; // they move too fast without dividing!
      radiusThenAngle[1] += radians(speedy / 3.0);
      if (band <= .5) {
        if (radiusThenAngle[0] > .5 && speedx > 0) {
          speedx = - speedx ;
          radiusThenAngle[0] = .5 ;
        }
      /*  } else if (band <= .67) {
        if (radiusThenAngle[0] > .67 && speedx > 0) {
          speedx = -speedx ;
          radiusThenAngle[0] = .67 ;
        } else if (radiusThenAngle[0] <= .33 && speedx < 0) {
          speedx = - speedx ;
          radiusThenAngle[0] = .335 ;
        } */
      } else {
        if (radiusThenAngle[0] > .99 && speedx > 0) {
          speedx = -speedx ;
          radiusThenAngle[0] = .99 ;
        } else if (radiusThenAngle[0] <= .6 && speedx < 0) {
          speedx = -speedx ;
          radiusThenAngle[0] = .65 ;
        }
      }
      float [] xThenY = polarToCartesian(radiusThenAngle[0], radiusThenAngle[1]);
      x = xThenY[0] * radius ;
      y = xThenY[1] * -radius ;
    } else { 
      x += speedx ;
      y += speedy ;
      if (x > domeWidthHeight/2) {
        x = domeWidthHeight /2 - 1 ;
        speedx = - speedx ;
      } else if (x < -domeWidthHeight /2) {
        x = (- domeWidthHeight /2) + 1 ;
        speedx = - speedx ;
      }
      if (y > domeWidthHeight/2) {
        y = domeWidthHeight/2 - 1 ;
        speedy = - speedy ;
      } else if (y < -domeWidthHeight/2) {
        y = (-domeWidthHeight/2) + 1 ;
        speedy = - speedy ;
      }
    }
    // linelen keeping inside clipping circle added 8/2018
    float linelen = dist(0, 0, x, y);
    if (linelen >= radius) {
      if (x == 0) {
        // center, just look at y
        if (y <= -radius) {
          y = -radius + 1 ;
          speedy = abs(speedy);
        } else if (y >= radius) {
          y = radius - 1 ;
          speedy = - abs(speedy);
        }
      } else if (y == 0) {
        if (x <= -radius) {
          x = -radius + 1 ;
          speedx = abs(speedx);
        } else if (x >= radius) {
          x = radius - 1 ;
          speedx = - abs(speedx);
        }
      } else {
        float scaledown = /*float*/(radius) / linelen;
        if (x > 0) {
          speedx = - abs(speedx);
        } else {
          speedx = abs(speedx);
        }
        if (y > 0) {
          speedy = - abs(speedy);
        } else {
          speedy = abs(speedy);
        }
        x = round(scaledown * x);
        y = round(scaledown * y);
        // println("DEBUG scaling line back by " + scaledown);
      }
    }
    if (rotateSpeedRadians != 0.0) {
      rotateRadians += rotateSpeedRadians ;
      while (rotateRadians >= TWO_PI) {
        rotateRadians -= TWO_PI ;
      }
      while (rotateRadians <= -TWO_PI) {
        rotateRadians += TWO_PI;
      }
    }
    if (shearxSpeedRadians != 0.0) {
      shearxRadians += shearxSpeedRadians ;
      while (shearxRadians >= TWO_PI) {
        shearxRadians -= TWO_PI ;
      }
      while (shearxRadians <= -TWO_PI) {
        shearxRadians += TWO_PI;
      }
    }
    if (shearySpeedRadians != 0.0) {
      shearyRadians += shearySpeedRadians ;
      while (shearyRadians >= TWO_PI) {
        shearyRadians -= TWO_PI ;
      }
      while (shearyRadians <= -TWO_PI) {
        shearyRadians += TWO_PI;
      }
    }
    if (scalexSpeed != 0) {
      scalex += scalexSpeed ;
      if (scalex > 2.0 || scalex < -2.0) {
        scalexSpeed = - scalexSpeed ;
      }
    }
    if (scaleySpeed != 0) {
      scaley += scaleySpeed ;
      if (scaley > 2.0 || scaley < -2.0) {
        scaleySpeed = - scaleySpeed ;
      }
    }
    if (spawnDegree > 1 || spawnDegree < -1) {
      // negative keeps hue constant, positive varies it.
      int absDegree = abs(spawnDegree);
      float newspeed = max(abs(speedx), abs(speedy));
      if (newspeed == 0) {
        newspeed = speedlow ;
      }
      int row = (absDegree-2) % spawnvectors.length ;
      float [][] myrow = spawnvectors[row];
      float huedeg = 360.0 / (absDegree+1) ;
      float scaler = 0.75; // / absDegree ;
      for (int f = 1, ix = 0 ; f <= absDegree ; f++, ix = (ix+1) % myrow.length) {
        ShapeData chld = clone();
        chld.myname = myname + "_" + spawnSerialNo++;
        chld.speedx = newspeed * myrow[ix][0];
        chld.speedy = newspeed * myrow[ix][1];
        chld.scalex = scalex * scaler ;
        chld.scaley = scaley * scaler ;
        if (spawnDegree > 0) {
          chld.hue = round(hue + f * huedeg) % 360 ;
        } else {
          chld.hue = hue ;
        }
        visible.put(chld.myname, chld);
      }
      if (! myname.contains("_")) {
        speedx = speedy = 0 ; // stall grandparentparent until children die
      }
    }
  }
  float [][][] spawnvectors = {
        {{-1.33, 1}, {1, -1.33}},  // two-way split
        {{-1.33, 1}, {1, 1.33}, {1, -1.33}},
        {{-1.33, 1}, {1, 1.33}, {1.33, 1}, {1, -1.33}},
        {{-1, 1.33}, {1.33, 1.5}, {-1.5, 1.33}, {1, -1.5}, {1.5, 1.33}},
        {{-1, 1.33}, {1.33, 1.5}, {-1.5, 1.33}, {1, -1.5}, {1.5, 1.33}, {-1.5, 1}},
  };
  /**
   * Call this *AFTER* calling pushMatrix() at top of display()
   * to do the translate, rotate, scale, shearX, shearY for the figure.
   * It sets colorMode(HSB,360,100,100,100);
   * It tests for 0 (1 for scale) before calling those functions.
   **/
  protected void helpDisplayTranslateRotateScaleShear() {
    colorMode(HSB, 360, 100, 100, 100);
    if (showfill) {
      tint(hue, saturation, brightness);
      fill(hue, saturation, brightness);
    } else {
      noFill();
      noTint();
    }
    if (showstroke || ! showfill) {
      stroke((hue+180)%360, saturation, brightness);
      strokeWeight(2);
    } else {
      noStroke();
    }
    translate(round(x), round(y));
    //println("TTT " + x + "," + y + " width = " + width + " height = " + height + " speedx " + speedx + " speedy " + speedy);
    if (rotateRadians != 0.0) {
      rotate(rotateRadians);
    }
    if (scalex != 1.0 || scaley != 1.0) {
      scale(scalex, scaley);
    }
    if (shearxRadians != 0.0) {
      shearX(shearxRadians);
    }
    if (shearyRadians != 0.0) {
      shearY(shearyRadians);
    }
  }
  /**
   * Define display() in subclass, where display() must call
   * pushMatrix() as first action, normally calls
   * helpDisplayTranslateRotateScaleShear second, and calls matching
   * popMatrix() just before returning.
   **/
  public abstract void display();
};

/**
 * Load an optional image file as a PImage. If the constructors PImage path
 * String is null, then a subclass draws a library shape instead.
 **/
class ImageDrawer extends ShapeData implements ShapeObjectTemplate {
  PImage img ;
  boolean cycleColors ;
  int lastFrame = -1 ;  // Added 2/16/2018 for pixels[] snapshot for 'f'
  public ShapeData clone() {
    ImageDrawer thing = new ImageDrawer(myname, hue);
    clonehelp(thing);
    thing.img = img ; // May need conditional img.copy() later.
    thing.cycleColors = cycleColors ;
    thing.saturation = saturation ;
    return thing ;
  }  
  public ImageDrawer(String myname, int myhue, String imageFile, 
    float rotspeed, float sclXspeed, float sclYspeed, 
    float shXSpeed, float shYSpeed, boolean cycleColors) {
    super(myname,myhue);
    if (imageFile != null) {
      img = loadImage(imageFile);
    }
    this.cycleColors = cycleColors ;
    shuffle();
    hue = myhue ; // override first call to shuffle for hue.
    float scaler = random(2.0, 5.0);
    if (img != null) {
      if (img.height > img.width) {
        scaler = height / scaler ; // pixels per displayed height
        scaley = scalex = scaler / img.height;
      } else {
        scaler = width / scaler ;
        scaley = scalex = scaler / img.width ;
      }
    }
    rotateSpeedRadians = rotspeed ;
    shearxSpeedRadians = shXSpeed ;
    shearySpeedRadians = shYSpeed ;
    scalexSpeed = sclXspeed ;
    scaleySpeed = sclYspeed ;
  }
  private ImageDrawer(String myname, int hue) {
    // Usedprivately by clone().
    super(myname, hue);
  }
  public void shuffle() {
    super.shuffle();
    super.move();
    //if (cycleColors) {
    /* DO NOT BLACKEN THE PIMAGE
     if (saturation == 0.0) { // black or white
     brightness = ((brightness == 0) ? 99 : 0);
     } else {
     hue = (hue + (int)random(20,300)) % 360 ;
     }
     */
    //}
    /* SCALING UP SWAMPS THE DOME!
    float scaler = random(2.0, 5.0);
    if (img != null) {
      if (img.height > img.width) {
        scaler = height / scaler ; // pixels per displayed height
        scaley = scalex = scaler / img.height;
      } else {
        scaler = width / scaler ;
        scaley = scalex = scaler / img.width ;
      }
    }
    */
  }
  public void move() {
    super.move();
    if (cycleColors && saturation != 0.0) {
      hue = (hue + 1) % 360 ;
    }
  }
  public void display() {
    pushMatrix();
    helpDisplayTranslateRotateScaleShear();
    if (img != null) {
      //println("HSB " + hue + "," + saturation + "," + brightness);
      image(img, 0, 0);
    }
    popMatrix();
  }
}

/**
 * Using ImageDrawer as a base helper class without a PShape image file,
 * draw a shape given by a non-negative shape index.
 **/
class ShapeDrawer extends ImageDrawer implements ShapeObjectTemplate {
  protected int myshape = 0 ;
  protected PShape svg ;
  public ShapeData clone() {
    ShapeDrawer thing = new ShapeDrawer(myname, hue);
    clonehelp(thing);
    thing.myshape = myshape ; // May need conditional img.copy() later.
    thing.svg = svg ;
    thing.img = img ; // May need conditional img.copy() later.
    thing.cycleColors = cycleColors ;
    thing.saturation = saturation ;
    return thing ;
  }
  public ShapeDrawer(String myname, int myhue, int myshape, 
    float rotspeed, float sclXspeed, float sclYspeed, 
    float shXSpeed, float shYSpeed, boolean cycleColors) {
    super(myname, myhue, null, rotspeed, sclXspeed, sclYspeed, 
      shXSpeed, shYSpeed, cycleColors);
    if (myshape >= 0) {
      this.myshape = myshape % 36 ; // See display() below for shapes.
      if (this.myshape == 3) {
        svg = loadShape("atom0e.0.svg");
      }
      // 6.. 13 are I Ching trigrams from 3-yin to 3-yang
      // 14.. 19 are dandelion puffs. Both use strokes only.
      // 20.. 25 are concentric circles. Both use strokes only.
    }
    //println("CONS SHAPE " + this.myshape);
  }
  private ShapeDrawer(String myname, int hue) {
    // Usedprivately by clone().
    super(myname, hue);
  }
  public void move() {
    // Added August 2018 to send MIDI x, y location to music generator.
    super.move();
    locateMIDI();          
  }
  public void shuffle() {
    // Added August 2018 to send MIDI x, y location to music generator.
    super.shuffle() ;
    locateMIDI();
  }
  protected void locateMIDI() {
    // Normalize to unit circle (clipped dome) so it hits the edge
    // of the Max Surround Panner exactly when it goes beyond the clipping
    // circle of the dome.
    if (visible.contains(this)) {
      // println("DEBUG locateMIDI() x,y = " + x + "," + y);
      // 0,0 is in the center per draw(), with Y going negative as you go up.
      float radius = min(width, height) / 2.0 ; // -radius TO +radius across the center
      float xfrom0 = ((-x) + radius) / 2.0 ; // 0..radius
      float yfrom0 = ((-y) + radius) / 2.0 ; // 0 at bottom, radius at top
      // The negative -y seems counter-intuitive, but it is here because
      // Live Max Surround orders speakers clockwise looking down at the floor,
      // while the visuals order speakers counterclockwise when looking up
      // at the dome while lying on your back.
      int xmidi = constrain(round(((xfrom0) / radius) * 128.0), 0, 127);
      int ymidi = constrain(round(((yfrom0) / radius) * 128.0), 0, 127);
      sendMIDI(CONTROLLER, myshape % NUMINSTRUMENTS, CTRLNUM+0, xmidi);
      sendMIDI(CONTROLLER, myshape % NUMINSTRUMENTS, CTRLNUM+1, ymidi);
    }
  }
  public void display() {
    rectMode(CENTER);
    ellipseMode(CENTER);
    pushMatrix();
    helpDisplayTranslateRotateScaleShear();
    int h5 = height/5 ;
    int h10 = height/10 ;
    int h20 = height/20 ;
    // println("DIS SHAPE " + myshape + "," + hue + "," + h20 + "@" + x + "," + y);
    switch (myshape) {
    case 0:
      ellipse(0, 0, h10, h10);
      //translate(0,0,-h10);
      //sphere(h10/4);
      break ;
    case 1:
      rect(0, 0, h10, h10);
      //box(h10/4);
      break ;
      /*
    case 0:
       ellipse(0, 0, h10, h10);
       break ;
       case 1:
       rect(0, 0, h10, h10);
       break ;
       */
    case 2:
      triangle(-h20, h20, h20, h20, 0, -h20);
      /*
      int degrees = 20 ;
       float rad = radians(degrees);
       pushMatrix();
       for (int xang = 0 ; xang < 360 ; xang += degrees) {
       rotateX(rad);
       for (int yang = 0 ; yang < 360 ; yang += degrees) {
       rotateX(rad);
       for (int zang = 0 ; zang < 360 ; zang += degrees) {
       rotateZ(rad);
       line(0,0,100,0);
       }
       }
       }
       popMatrix();
       */
      break ;
    case 4: // Swap 3 & 4 Aug. 2018 take make the hexagon come earlier
      quad(0, -h10, h20, 0, 0, h10, -h20, 0);
      break ;
    case 3:
      svg.setTint(color(hue, 100, 100, 100)); // hexagon
      if (showfill) {
        svg.setFill(color(hue, 100, 100, 100));
      } else {
        svg.setFill(color(hue, 100, 100, 0));
      }
      svg.setStroke(color((hue+180)%360, 100, 100, 100));
      shape(svg, 0, 0, h10, h10);
      break ;
    case 5: // recursive snapshot of 2D display on 'f' 'F'
    // arc(0, 0, h10, h10, 0, PI+QUARTER_PI, PIE); // OLD pacman arc
      if (lastFrame < 0
        || ((! IamAclone) && (lastFrame != frameCount))) {
        // an aging clone does not change shape
        lastFrame = frameCount ;
        img = createImage(width, height, ARGB);
        img.loadPixels();
        loadPixels();
        System.arraycopy(pixels, 0, img.pixels, 0, pixels.length);
        // Update August 12-13, 2018 to use alphamask to blank
        // img pixels outside the central circle.
        int [] alphamask = new int [width*height];
        int h2 = height / 2, w2 = width / 2 ;
        int radius = min(width, height) / 2 ;
        if (width > height) {
          int colstart = (width-height) / 2 ;
          int colend = colstart + height ;
          for (int col = colstart; col < colend ; col++) {
            for (int row = 0; row < height; row++) {
              if (dist(col,row,w2,h2) <= radius
                  /*&& ((img.pixels[row * width + col] & 0x0ffffff) != 0)*/) {
                alphamask[row * width + col] = 0x0ffffffff ;
              }
            }
          }
        } else {
          int rowstart = (height-width) / 2 ;
          int rowend = rowstart + width ;
          for (int col = 0; col < width ; col++) {
            for (int row = rowstart; row < rowend ; row++) {
              if (dist(col,row,w2,h2) <= radius) {
                alphamask[row * width + col] = 0x0ffffffff ;
              }
            }
          }
        }
        img.updatePixels();
        img.mask(alphamask);
        img.resize(h5, h5);
      }
      imageMode(CENTER);
      noTint();
      image(img, 0, 0, h5, h5); // h10 size is experimental
      break;
    default:
      strokeWeight(constrain(h20/20, 2, 6));
      if (myshape >= 6 && myshape <= 13) {
        int trigram = myshape - 6 ;
        for (int t = 0; t < 3; t++, trigram = trigram >> 1) {
          if ((trigram & 1) == 0) {
            // yin at this spot
            line(0, t*h20, h10/3, t*h20);
            line(2*h10/3, t*h20, h10, t*h20);
          } else {
            // yang
            line(0, t*h20, h10, t*h20);
          }
        }
      } else if (myshape >= 14 && myshape <= 19) {
        int rays = myshape - 12 ;
        for (int i = 0; i < rays; i++) {
          pushMatrix();
          translate(h20, h20);
          rotate(radians(i*180.0/rays));
          translate(-h20, -h20);
          line(h20, 0, h20, h10);
          popMatrix();
        }
      } else if (myshape >= 20 && myshape <= 25) {
        int rays = myshape - 18 ;
        noFill();
        for (int i = 0; i < rays; i++) {
          ellipse(0, 0, h20*(i+1)/2, h20*(i+1)/3);
        }
      } else if (myshape >= 26 && myshape <= 30) {
        int rays = myshape - 24 ;
        noFill();
        for (int i = 0; i < rays; i++) {
          rect(0, 0, h20*(i+1)/2, h20*(i+1)/3);
        }
      } else if (myshape >= 31 && myshape <= 35) {
        int rays = myshape - 29 ;
        noFill();
        for (int i = 0; i < rays; i++) {
          pushMatrix();
          scale(rays/4.0/(i*2.0+1.0));
          line(-h10, -h20, h10, -h20);
          line(-h10, h20, h10, h20);
          line(0, -h20, 0, h20);
          popMatrix();
        }
      } else {
        ellipse(0, 0, h10/2, h10/3);
      }
      strokeWeight(2);
      break ;
    }
    popMatrix();
  }
}


// See documentation at the top of this sketch code
// to see the keyboard commands interpreted in this function.
void keyPressed() {
  // println("DEBUG key INT " + int(key) + ", " + key);
  if (key == CODED) {
    if (keyCode == DOWN) {
      priorScale = priorScale - .005 ;
      if (priorScale <= 0.0) {
        priorScale = .1 ;
      }
      println("priorScale = " + priorScale);
    } else if (keyCode == UP) {
      priorScale = priorScale + .005 ;
      if (priorScale > 2.0) {
        priorScale = 2.0 ;
      }
      println("priorScale = " + priorScale);
    } else if (keyCode == RIGHT) {
      float pr = (degrees(priorRotate) + 1.0);
      while (pr >= 360.0) {
        pr -= 360.0 ;
      }
      priorRotate = radians(pr);
      println("priorRotate = " + pr);
    } else if (keyCode == LEFT) {
      float pr = (degrees(priorRotate) - 1.0);
      while (pr <= -360.0) {
        pr += 360.0 ;
      }
      priorRotate = radians(pr);
      println("priorRotate = " + pr);
    }
    return ;
  }
  if (key == 'F'&& ! isCommandMode()) {
    isFrozen = ! isFrozen ;
    println("Frozen? " + isFrozen);
  } else if (key == 'f' && ! isCommandMode()) {
    showfill = ! showfill ;
    println("showfill? " + showfill);
  } else if (key == 'k' && ! isCommandMode()) {
    showstroke = ! showstroke ;
    println("showstroke? " + showstroke);
  } else if (key == 'C' && ! isCommandMode()) {
    // avoid collision with the +C to add an object of class c
    isClear = true ;
  } else if (key == 'M' && ! isCommandMode()) {
    tomove = ! tomove ;
  } else if (key == '|') {
    flipx = - flipx ;
  } else if (key == '_') {
    flipy = -flipy ;
  } else if (key == '/' || key == '\\') {
    flipx = flipy = 1 ;
  } else if (key == 'R' && ! isCommandMode()) {
    doReset();
  } else if (key == 'l' && ! isCommandMode()) {
    meltindex = 2 ;
    nextcapture = lastcapture = 0 ;
    meltBothWays = false ;
  } else if (key == 'L' && ! isCommandMode()) {
    meltindex = -2 ;
    nextcapture = lastcapture = 0 ;
    meltBothWays = false ;
  } else if (key == 'm' && ! isCommandMode()) {
    meltindex = 2 ;
    nextcapture = lastcapture = 0 ;
    meltBothWays = true ;
    // println("DEBUG SETTING meltBothWays true!!");
  } else if (key == 'V' && ! isCommandMode()) { // Add one of each keyed objects.
    for (String k : team.keySet()) {
      visible.put(k, team.get(k));
    }
  } else if (key == 'v' && ! isCommandMode()) { // Remove objects, let history clear out.
    visible.clear();
    // history.clear();
  /*
  } else if (key == '(') {
    cliptype = 2 ; // circle
  } else if (key == ')') {
    cliptype = 3 ; // ellipse
  } else if (key == '[') {
    cliptype = 1 ; // square
  } else if (key == ']') {
    cliptype = 0 ; // noclip
  */
  } else if (key == '{') {
    hashX = ! hashX ;
    hashXOffset = 0 ;
  } else if (key == '}') {
    hashY = ! hashY ;
    hashYOffset = 0 ;
  } else if (key == '(') {
    priorRotate = 0.0 ;
    priorShearX = 0 ;
    priorShearY = 0 ;
    priorShearXSpeed = 0 ;
    priorShearYSpeed = 0 ;
    flipx = flipy = 1 ;
    hashX = hashY = false ;
    hashXOffset = hashYOffset = 0 ;
  } else if (key == ')') {
    priorScale = 1.0 ;
    priorShearX = 0 ;
    priorShearY = 0 ;
    priorShearXSpeed = 0 ;
    priorShearYSpeed = 0 ;
    flipx = flipy = 1 ;
    hashX = hashY = false ;
    hashXOffset = hashYOffset = 0 ;
  } else if (key == '[') {
    priorTranslateX = 0 ;
    priorShearX = 0 ;
    priorShearY = 0 ;
    priorShearXSpeed = 0 ;
    priorShearYSpeed = 0 ;
    flipx = flipy = 1 ;
    hashX = hashY = false ;
    hashXOffset = hashYOffset = 0 ;
  } else if (key == ']') {
    priorTranslateY = 0 ;
    priorShearX = 0 ;
    priorShearY = 0 ;
    priorShearXSpeed = 0 ;
    priorShearYSpeed = 0 ;
    // flipx = flipy = 1 ;
    hashX = hashY = false ;
    hashXOffset = hashYOffset = 0 ;
  } else if (key == '#') {
    priorTranslateX = 0 ;
    priorShearX = 0 ;
    priorShearY = 0 ;
    flipx = flipy = 1 ;
    priorTranslateY = 0 ;
    priorShearXSpeed = 0 ;
    priorShearYSpeed = 0 ;
    hashX = hashY = false ;
    hashXOffset = hashYOffset = 0 ;
  } else if (key == '+') {
    command = "+";
  } else if (key == '-' && command.length() == 0) {
    command = "-";
  } else if (key == 'x' && ! isCommandMode()) {
    command = "x";
  } else if (key == 'y' && ! isCommandMode()) {
    command = "y";
  } else if (key == 'X' && ! isCommandMode()) {
    command = "X";
  } else if (key == 'Y' && ! isCommandMode()) {
    command = "Y";
  } else if (key == 'u' && ! isCommandMode()) {
    command = "u";
  } else if (key == 'w' && ! isCommandMode()) {
    command = "w";
  } else if (key == 'W' && ! isCommandMode()) {
    command = "W";
  } else if (key == 'Z' && ! isCommandMode()) {
    command = "Z";
  } else if (key == 'r' && ! isCommandMode()) {
    command = "r";
  } else if (key == 's' && ! isCommandMode()) {
    command = "s";
  } else if (key == 'B' && ! isCommandMode()) {
    command = "B";
  } else if (key == '@') {
    command = "@";
  } else if (key == 'e' && ! isCommandMode()) {
    command = "e";
  } else if (key == 'E' && ! isCommandMode()) {
    command = "E";
  } else if (key == '~') {
    doShuffle();
  } else if (key == 'G' && ! isCommandMode()) {
    isgray = ! isgray ;
  } else if (key == 'N' && ! isCommandMode()) {
    isinvert = ! isinvert ;
  } else if (key == 'c' && ! isCommandMode()) {
    isPolar = false ;
  } else if (key == 'p' && ! isCommandMode()) {
    isPolar = true ;
  } else if (key == 'O' && ! isCommandMode()) {
    iserode = ! iserode ;
    if (iserode) {
      isdilate = false ;
    }
  } else if (key == 'D' && ! isCommandMode()) {
    isdilate = ! isdilate ;
    if (isdilate) {
      iserode = false ;
    }
  } else if (key == 'P' && ! isCommandMode()) {
    command = "P";
  } else if (key == 'T' && ! isCommandMode()) {
    command = "T";
  } else if (key == 't' && ! isCommandMode()) {
    command = "t";
  } else if (key == 'g' && ! isCommandMode()) {
    command = "g";
  } else if ((key == 'i' || key == 'I') && ! isCommandMode()) {
    command = "" + key ;
    meltBothWays = false ;
  } else if (key == '^' && ! isCommandMode()) {
    command = "^";
  } else if ((command.startsWith("r") || command.startsWith("s")
    || command.startsWith("@") || command.startsWith("e") || command.startsWith("E")
    || command.startsWith("x") || command.startsWith("y")
    || command.startsWith("X") || command.startsWith("Y")
    || command.startsWith("u") || command.startsWith("w") || command.startsWith("W")
    || command.startsWith("t") || command.startsWith("g") || command.startsWith("Z")
    || command.startsWith("P") || command.startsWith("T")
    || command.startsWith("B") || command.startsWith("^")
    || command.startsWith("i") || command.startsWith("I")
    )
    && ((key >= '0' && key <= '9') || (key == '.') || (key == '-')
    || (key == '\n'))) {
    if (key != '\n') {
      command = command + key ;
    } else {
      try {
        println("DEBUG key " + key + ", command " + command);
        float v = Float.parseFloat(command.substring(1));
        
        if (command.startsWith("r")) {
          priorRotate = radians(v);
          println("priorRotate = " + v);
        } else if (command.startsWith("s")) {
          priorScale = v ;
          println("priorScale = " + priorScale);
        } else if (command.startsWith("x")) {
          float percent = constrain(v, -1.0, 1.0);
          priorTranslateX = round(percent * width) ;
          // println("v is " + v + ", SET priorTranslateX to " + priorTranslateX);
        } else if (command.startsWith("y")) {
          float percent = constrain(v, -100.0, 100.0);
          priorTranslateY = round(percent / 100.0 * height) ;
          println("v is " + v + ", SET priorTranslateY to " + priorTranslateY);
        } else if (command.startsWith("X")) {
          priorShearX = constrain(v, -360.0, 360.01);
        } else if (command.startsWith("Y")) {
          priorShearY = constrain(v, -360.0, 360.01);
        } else if (command.startsWith("u")) {
          priorShearXSpeed = constrain(v, -360.0, 360.01);
        } else if (command.startsWith("w")) {
          priorShearYSpeed = constrain(v, -360.0, 360.01);
        } else if (command.startsWith("W")) {
          wheelCount = constrain(v, 0.0, 16.0);
          if (wheelCount == 0) {
            wheelCount = 1 ;
          } else if (wheelCount > 1.0) {
            wheelCount = round(wheelCount);
          }
          println("DEBUG wheelCount " + wheelCount);
        } else if (command.startsWith("Z")) {
          wheelScale = constrain(v, .1, .9);
          println("DEBUG wheelScale " + wheelScale);
        } else if (command.startsWith("@")) {
          int rate = constrain(round(v), 1, 120);
          // double oldrate = frameRate ;
          frameRate(rate);
        } else if (command.startsWith("e")) {
          eraseRate = constrain(round(v), 0, 255);
          println("DEBUG eraseRate " + eraseRate);
        } else if (command.startsWith("E")) {
          colorMode(HSB,360,100,100,100);
          int tmpc = constrain(round(v), -1, 400);
          if (tmpc <= 0) {
            // use 1 if you want red
            eraseColor = color(0,0,0,100);
          } else if (tmpc > 359) {
            eraseColor = color(0,0,100,100); // white
          } else {
            eraseColor = color(tmpc, 100, 100, 100);
          }
          //background(eraseColor);
          println("DEBUG eraseColor " + eraseColor);
        } else if (command.startsWith("t")) {
          tailLength = constrain(round(v), 0, 15);
          if (tailLength == 0) {
            history.clear();
          }
          println("DEBUG tailLength " + tailLength);
        } else if (command.startsWith("g")) {
          tailGap = constrain(round(v), 1, 50);
          println("DEBUG tailGap " + tailGap);
        } else if (command.startsWith("P")) {
          posterizeLevel = constrain(round(v), 0, 255);
          println("DEBUG posterizeLevel " + posterizeLevel);
        } else if (command.startsWith("T")) {
          thresholdLevel = constrain(v, 0.0, 1.0);
          println("DEBUG thresholdLevel " + thresholdLevel);
        } else if (command.startsWith("B")) {
          blendState = int(constrain(v, 0, blendModes.length-1));
          println("DEBUG blendState " + blendState);
        } else if (command.startsWith("i") || command.startsWith("I")) {
          nextcapture = constrain(v, -128, 128) ;
          captureShape = command.charAt(0);
          if (meltmaskErode != null) {
            meltmaskErode = null ;
            captureimg = null ;
            meltmaskDilate = null ;
          }
          meltindex = 0 ;
          println("set nextcapture to " + captureShape + nextcapture);
        } else if (command.startsWith("^")) {
          spawnDegree = round(constrain(v, -6.1, 6.1));
          println("DEBUG spawnDegree " + spawnDegree);
          if (abs(spawnDegree) == 1) {
            String [] ksary = visible.keySet().toArray(new String [0]);
            // Added next Jan2019 for objects spawned from a parent object.
            // Iterate over array copy so to avoid concurrent mutation.
            for (String ks : ksary) {
              if (ks.contains("_")) {
                visible.remove(ks);
              }
            }
            spawnDegree = 0 ;
          }
          
        }
        command = "";
      } 
      catch (NumberFormatException fx) {
        command = "";
      } 
      
    }
  } else if ((command.startsWith("+") || command.startsWith("-"))
    && ((key >= 'a' && key <= 'z') || (key >= 'A' && key <= 'Z')
    || (key >= '0' && key <= '9'))) {
    command = command + key ;
    if (command.length() == 2) {
      String className = command.substring(1);
      if (command.startsWith("-")) {
        visible.remove(className);
        // println("REMOVING " + className);
        String [] ksary = visible.keySet().toArray(new String [0]);
        className = className + "_" ;
        // Added next Jan2019 for objects spawned from a parent object.
        // Iterate over array copy so to avoid concurrent mutation.
        for (String ks : ksary) {
          if (ks.startsWith(className)) {
            visible.remove(ks);
            // println("REMOVING " + ks);
          }
        }
      } else {
        ShapeData shape = team.get(className);
        if (shape != null) {
          shape.shuffle();
          visible.put(className, shape);
          // println("PUTTING " + className);
        }
      }
      command = "";
    }
  }
}

void mousePressed() { // Save coordinates. All work happens in mouseReleased().
  pressedX = mouseX ;
  pressedY = mouseY ;
}

void mouseReleased() {
  if (mouseX < 100 && mouseY < 100) {
    return ;  // top left corner used to grab focus
  }
  if (abs(mouseX-pressedX) > 10 || abs(mouseY-pressedY) > 10) {
    // shearX or shearY
    float sx = (float)(mouseX-pressedX)  / width ;
    float sy = (float)(mouseY-pressedY)  / height ;
    if (sx > sy) {
      priorShearX = sx * 360.0 ;
    } else {
      priorShearY = sy * 360.0 ;
    }
    return ;
  }
  ///* IGNORE MOST DOME CLICKS SO WE CAN GET FOCUS WITHOUT SCALE/ROTATE CHANGES
  priorRotate = TWO_PI * (float) mouseX / width ;
  println("priorRotate = " + degrees(priorRotate));
  if (mouseY >= (height-1)) {
    priorScale = 1.0;
  } else {
    priorScale = 2.0 * (float) mouseY / height ;
  }
  println("priorScale = " + priorScale);
  if (mouseX <= 10) { // Left margin cancels out the following:
    priorShearX = 0 ;
    priorShearY = 0 ;
    priorTranslateX = 0 ;
    priorTranslateY = 0 ;
  }
  //*/
}

boolean isCommandMode() {
  return (("+".equals(command) || "-".equals(command)));
}

float distancefromCenter(int x, int y) {
  return sqrt(x * x + y * y);
}

float distanceFromHere(int herex, int herey, int therex, int therey) {
  return(distancefromCenter(herex-therex, herey-therey));
}

synchronized void doReset() {
  isReset = true ;
  isClear = true ;
  showfill = true ;
  isgray = iserode = isdilate = isinvert = false ;
  visible.clear(); // Remove all objects from the map-to-plot.
  history.clear();
  tailLength = 0 ;
  priorRotate = 0.0 ;
  priorScale = 1.0 ;
  priorTranslateX = 0 ;
  priorTranslateY = 0 ;
  priorShearX = 0 ;
  priorShearY = 0 ;
  priorShearXSpeed = 0 ;
  priorShearYSpeed = 0 ;
  hashXOffset = hashYOffset = 0 ;
  wheelCount = 1 ;
  wheelScale = .5 ;
  flipx = 1 ;
  flipy = 1 ;
  eraseRate = 255 ; // 0 ;
  blendState = 0 ;
  tomove = true ;
  nextcapture = lastcapture = 0 ;
  captureimg = null ;
  meltmaskErode = null ;
  meltmaskDilate = null ;
  meltindex = meltlowpixel = 0 ;
  meltBothWays = false ;
}

synchronized void doShuffle() {
  for (ShapeObjectTemplate thingy : visible.values()) {
    thingy.shuffle();
  }
}

// See http://midi.teragonaudio.com/tech/midispec.htm
// and https://www.midi.org/specifications/item/gm-level-1-sound-set
// See Parson's code at http://faculty.kutztown.edu/parson/spring2017/MidiKBSpring2016Parson.txt
void sendMIDI(int opcode, int chan, int data1, int data2) {
  if (receiver == null) {
    return ;
  }
  //println("DEBUG ENTER sendMIDI " + opcode + "," + chan + "," + data1 + "," + data2);
  try {
    ShortMessage noteMessage = new ShortMessage() ;
    noteMessage.setMessage(opcode, chan, data1, data2);
    receiver.send(noteMessage, -1L);  // send it now
  } 
  catch (InvalidMidiDataException dx) {
    println("MIDI ERROR: " + dx.getMessage()); // Error messages go here.
  }
  //println("DEBUG EXIT sendMIDI");
}
